/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
     author: kekec 		E-mail: behaviour.brain.research@gmail.com
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          lpt-2 version: 1.0


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: Total: 20 options

var c_min = 8;  // Minimum number of generated characters (columns)

var c_max = 10; // Maximum number of generated characters (columns)

var iter = 1;   // Number of iterations for each row value

var stype = true;   // Solution type - accepted: Complete ('1') = true or Partial ('2') = false

var is_lower = '3'; // Letter type: '1' = Original char set from text file, '2' = lowercase; '3' = UPPERCASE

var is_ltr = '1';   // Parsing (reading) direction: '1' = left-to-right, '2' = right-to-left, '3' = both. see getScore()

var limen = 2;  // Minimum word length to be accepted as valid (threshold)

var timer_enabled = false;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 600;     // Task time in seconds if clock = 1000ms (default). 

var font_family = 'Arial, Helvetica, sans-serif';  // default canvas ctx.fontFamily 

var is_back = true;    // if (is_back==true) => user can go back to previous task/item. (true || false)

var prg_type = '0';   //  type of information in Progress Bar ('0' => Items, '1' => Time; '2' => Nothing) 

var is_shuffle = false;   // Randomly shuffle rows (only when Demo mode = Off)

var demo = true;       // if (demo == true) => Show task solutions 

var bclr_1 = '#fff'; // Background color (id="can_1")			
 
var pclr_1 = '#000';   // Font color  (id="can_1")

var bclr_2 = '#fffff0';  // Background color (for top line). default: Ivory. id="can_2". Used only if stype = false. Solution type-accepted => Partial ('2')

var pclr_2 = '#000';  // Font color (for top line). id="can_2". Used only if stype = false. Solution type-accepted => Partial ('2')

var is_border = false;   // Draw border around numbers

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 


// Other global variables:

const clock = 1000;  // basic timer resolution in ms

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var counter = 0;      // timer counter. if (counter >= time) Timer will call theEnd

var item_cnt;       // number of items in the task. range [0..item_cnt-1]. see setPar()

var item;   				// current item in task

var is_task = false;    // task in progress or not for pageResize() event. It is true for all events between end of startTask() and theEnd() events

var dirty = false;  	// if  (dirty==true) => user changed any options in form 

var is_storage = false;		// local storage is enabled on user browser or not
 
var wor_cnt;   			// Total number of words loaded from the dictionary. default: 370067

var dic_def = true;  // default Dictionary is used (true) or not (user loaded custom Dictionary loaded from txt file) 
 
 
var stim = [];   // stimulus array. Dimensions: [items] x [character count in this item]. see stimGen()   

var ans = [];    // user answer array. Dimensions: [items] x [character count in this item]. see stimGen()  

var cor = [];    // correct anwswers in task (note: can be one or more correct answers for each item: >=1). see stimGen()  


var rndKey = function (obj) {
    let keys = Object.keys(obj);
    return keys[ keys.length * Math.random() << 0];
};

/*
var rndValue = function (obj) {
    let keys = Object.keys(obj);
    return obj[keys[ keys.length * Math.random() << 0]];
}; */



/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */



// application init [create/start Timer, check storage, show start panel..]
function init() 
{ 
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);  
	
	is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }	
		
	loadFont();
	
	setIns();
	
	displayFreq();
	
	showStart(); 	
}


// application timer handler function
function appTimer()  
{    
	if (timer_enabled && is_task)   // else: nop, idle, do nothing 
	{
		counter++;  
		if (counter >= timer)  {     // task time is limited
			theEnd(); 
		}	
		else if (prg_type == '1')   {   // '1' => Time. display time on Progress Bar  
			let rest= timer-counter;   // this is remaining time 
			let percent= Math.round(100*rest/timer);    
			document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
			document.getElementById('nav_items').textContent= stringTime(rest);	
		}
	}  
}	


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  let hh = Math.floor(total/(60*60));         
  let div_min = total % (60 * 60);
  let mm = Math.floor(div_min / 60);       
  let div_sec = div_min % 60;
  let ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 (probably redundant)
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp	
	const arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	let sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement('option');
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex= 3; // 'Arial, Helvetica, sans-serif'
	document.getElementById('font_family_msg').style.fontFamily= sel.value; // = arr[3];
}	


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		

  document.getElementById('ins_stype').textContent= stype ? 'The screen will display the letters of the alphabet mixed in random order. Your task is to use the drag and drop action to move the letters to a position so that they form a valid word.' : 'The letters of the alphabet will be displayed at the bottom of the screen. Your task is to use the Drag and drop action to move these letters to the top line of the screen so that the selected letters form a valid word. The selected letters in the top line of the screen do not have to be stacked one after the other, ie if there is space on the screen you can leave spaces between the letters. You don’t have to use all the letters. If you want your word to be valid, the minimum number of letters used must be: ' + limen + '.';

	document.getElementById('ins_back').style.display= is_back ? 'none' : 'block';
	
	if ( document.getElementById('timer_enabled').checked )
  {   
    document.getElementById('ins_timer').setAttribute('style', 'display:block'); 		
		document.getElementById('ins_timer').textContent= 'The duration of this task is limited.';  // The duration of this task is always limited.
		if (prg_type == '0')  {
			document.getElementById('ins_timer').textContent+= ' The Progress bar will show how many items are left until the end of the task.';
		}
		else if (prg_type == '1')  {
			document.getElementById('ins_timer').textContent+= ' The Progress bar will display the remaining time until the end of the task.';  
		}
		else  {
			document.getElementById('ins_timer').textContent+= ' But information about the remaining time will not be visible.';  
		}  
  }
	else  {
		document.getElementById('ins_timer').setAttribute('style', 'display:none');
	}	

	document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';	
	
}	


// Show the loaded number of words for the currently used dictionary (grouped by word length).
// set id = c_min, c_max (Minimum number of letters used, Maximum number of letters used)
function displayFreq() 
{
	let cnt= Object.keys(words).length;  // default: 370067
	if (cnt == 0) {
		document.getElementById('freq_rep').innerHTML= 'Dictionary not loaded successfully. The total number of words is zero.';
    document.getElementById('words_cnt').textContent= '0';	
    wor_cnt= 0;
		document.getElementById('item_fs').style.display= 'none';  // hide c_min, c_max, iter, item_total		
	}
	else 
	{
	  let len = [], sum= 0;
    let s = 'Number of different word lengths loaded from the dictionary = ' + cnt + '<br>The number of words loaded (grouped by word length):';		
		for (let key in words)   {  // 1. key is word length
			let f = Object.keys( words[key] ).length;
			len.push(key);
			s+= '<br>Word length= ' + key + '. Number= ' + f;
			sum+= f;
		}
		document.getElementById('words_cnt').textContent= sum;  // Update Total number of all words loaded from the dictionary
		wor_cnt= sum;
		s+= '<br>Total number of words loaded from the dictionary =  ' + sum;
		document.getElementById('freq_rep').innerHTML= s;
	
		// must be at least one valid words and minimum 1 word with string length 2 
		if ( len.length == 0 || getMaxArray(len) < 2 )  {     // minimum loaded f_max must be >= 2
      document.getElementById('item_fs').style.display= 'none';  // hide c_min, c_max, iter, item_total	
		}			
		else  
		{
			document.getElementById('item_fs').style.display= 'block';  // show c_min, c_max, iter, item_total	
			let f_min= getMinArray(len), f_max= getMaxArray(len); 	
			let e_min= document.getElementById('c_min');  	
			if ( parseInt(e_min.value, 10) > f_max ) {
				e_min.value= 2;  
			}
			//if ( f_max >= 2 ) {   // already checked. minimum loaded f_max is always >= 2
				e_min.max= f_max;
			//}
			let e_max= document.getElementById('c_max');	
			if ( parseInt(e_max.value, 10) > f_max ) {
				e_max.value= f_max;
			}	
			//if ( f_max >= 2 ) {   // already checked. minimum loaded f_max is always >= 2
				e_max.max= f_max;  
			//}
			//document.getElementById('iter').max= f_min;   // depends on user selected range: c_min..c_max
		}
	}
}	


// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Math/max
// get the minimum of an numeric array
function getMinArray(numArray) {
  return Math.min.apply(null, numArray);
}


//  get the maximum of an array. This should only be used for arrays with relatively few elements
function getMaxArray(numArray) {
  return Math.max.apply(null, numArray);
}


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{ 
	if ( is_task )  { 
		document.getElementById('can_m').style.display= 'none';  // tmp until mouseMove()
		setCanvas();		
		drawStim();		
	}
}	


// show start panel 
function showStart()  
{     
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (is_storage && localStorage.getItem('lpt-2-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('lpt-2-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';  // xxx
		}
	}
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
	refMsg(); 
  setIns();	
}


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty = true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'show_f':
		  document.getElementById('freq_rep').style.display = e.value == '1' ? 'block' : 'none';
			break;
		
		case 'c_min':
		case 'c_max':
			if (e.validity.valid)  {
				msg.textContent= ''; 				
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
			} 
			checkItem();
			break; 
   	  
		case 'iter':
			if (e.validity.valid)  {
				msg.textContent= 'Number of repetitions for each length. Must be less than or equal to the smallest number of existing words loaded from the dictionary within the requested word length interval.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			checkItem();
			break; 	
			
    case 'stype':  
			if (e.value == '1')  {  // '1' => Complete
				msg.textContent=  'Letters are displayed in one line only (full screen).  The user needs to use the Drag and drop action to move the letters so that they form a word from the loaded dictionary. All displayed letters must be used.';
				document.getElementById('p_limen').style.display= 'none';
				document.getElementById('p_bclr_2').style.display= 'none';
				document.getElementById('p_pclr_2').style.display= 'none';
			}
			else  {
				msg.textContent= 'The letters are displayed at the bottom of the screen. The user should use the Drag and drop action to move the selected letters from the bottom of the screen to the top of the screen. Not all letters need to be selected. You can define a minimum threshold (number of letters) for this selection to be valid.';
				document.getElementById('p_limen').style.display= 'block';
				document.getElementById('p_bclr_2').style.display= 'block';
				document.getElementById('p_pclr_2').style.display= 'block';
		  }
			break;
		
		case 'is_lower':	
			switch ( e.value )	{
				case '1':
					msg.textContent= 'The text renders as it is. When using your own stimuli, probably the most reliable option is to choose the original.';
					break;
				case '2':
					msg.textContent= 'Lowercase characters will always be displayed on the screen.';
					break;	
				case '3':
					msg.textContent= 'UPPERCASE characters will always be displayed on the screen.';
					break;		
      }     
			break; 
				
		case 'is_ltr':		
		  switch ( e.value ) {
				case '1':
			    msg.textContent= 'Default. Left-to-right text direction. Change this option only if you do not use the default Dictionary, and if the reading direction is reversed in your language.';
					break;
			  case '2':
			    msg.textContent= 'Right-to-left text direction.';
					break;
				case '3':
			    msg.textContent= 'The original and reverse answers are equivalent. Eg if the user chooses the word "example" OR "elpmaxe" both are valid.';
					break;			       
			}
			//msg.textContent += ' Applies only to parsing text in the top line.';
			break; 
		
		case 'limen':
			if (e.validity.valid)  {
				msg.textContent= 'This is the minimum validity threshold. All participant answers that are shorter than the selected one will always be invalid (0 points), regardless of whether the word is in the dictionary or not.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 	
		
		case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes.';
        document.getElementById('p_timer').style.display= 'block';
      }
      else  {  
        msg.textContent= 'No. Timer is disabled.';
        document.getElementById('p_timer').style.display= 'none'; 
      }       
      break;
		
    case 'timer':
			if (e.validity.valid)  {
				msg.textContent= stringTime(e.value); 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 

		/* case 'font_family':
		  break; */
		
		case 'is_back':		
			msg.textContent= e.checked ? 'Yes.' : 'No.';      
			break; 
						
		case 'prg_type':                                                                                                                                            
      if (e.value == '0')  {
        msg.textContent= 'It will be displayed how many items are left to the end.'; // The number of the current item will be displayed, as well as the total number of items.
      }
      else if (e.value == '1')  {
        msg.textContent= 'The remaining time until the end of the task will be displayed. Only when timer is enabled.';
      }
      else  {
        msg.textContent= ''; //'Nothing';
      }
      break;    

    case 'is_shuffle':		
			msg.textContent= e.checked ? 'Yes. All stimuli will be randomly mixed. The ordinal number of the item and the length of the letters within the item will not be directly correlated.' : 'No. Stimuli with a smaller number of letters will be displayed first. Then the number of letters will increase up to the maximum length.';      
			break; 

		case 'demo': 
			msg.textContent= e.checked ? 'On. Default task solution will be displayed at the bottom of the screen. As there may be multiple solutions, you can check that your currently chosen solution is correct by clicking the Check button.' : 'Off';  
			break; 
	
		case 'bclr_1': 
		case 'pclr_1':
		case 'bclr_2': 
		case 'pclr_2':    
			msg.textContent= validColor(e.value)  ? '' : 'The selected color is invalid.'; 
			break; 	

    case 'is_border':		
			msg.textContent= e.checked ? 'Yes.' : 'No.';      
			break; 
		
    case 'newline':		
			document.getElementById('span_delim').style.display= e.checked ? 'none' : 'inline';      
			break; 			

		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// check if item count is valid or not
// update item count on options form and check c_min <= c_max and update item_total label
function checkItem()  
{
  let c_min = document.getElementById('c_min');
  let c_max = document.getElementById('c_max');
  let iter = document.getElementById('iter');
  let items = document.getElementById('item_total');
  let c_min_msg = document.getElementById('c_min_msg');
  let c_max_msg = document.getElementById('c_max_msg');    
  if (c_min.validity.valid && c_max.validity.valid && iter.validity.valid)   
  {                      
    if  ( parseInt(c_min.value,10) > parseInt(c_max.value,10) )  
    {  
			let tmp= 'Minimum number of letters must not be higher than the Maximum number of letters.';
      c_min_msg.textContent= tmp;  
      c_max_msg.textContent= tmp;  
      items.textContent= 'Total number of items = Invalid value'; 
      items.setAttribute('class', 'error');
    } 
    else   
    {                                                                                               
      c_min_msg.textContent= '';  
      c_max_msg.textContent= '';
      let cnt= 0, level= parseInt(c_min.value,10);  // update cnt => total item count in task  
      while (level <= parseInt(c_max.value,10))  {
        cnt+= parseInt(iter.value,10);    
        level++;   
      } 
      items.textContent=  'Total number of items = ' + cnt ;      
      items.removeAttribute('class');
    }
  } 
  else  
  {
		let s = 'Values within this range are allowed: ', s2= ' with step= ';    
    c_min_msg.textContent= s + c_min.min + ".." + c_min.max + s2 + c_min.step;
		c_max_msg.textContent= s + c_max.min + ".." + c_max.max + s2 + c_max.step;
		items.textContent= 'Total number of items = Invalid value'; 
    items.setAttribute('class', 'error');
  }    
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity most of options. for other user options: see setPar()
function optionsValid()  
{ 
  let res = document.getElementById('c_min').validity.valid 
				 && document.getElementById('c_max').validity.valid
	       && document.getElementById('iter').validity.valid 			 
				 && document.getElementById('limen').validity.valid 
				 && document.getElementById('timer').validity.valid 
				 && validColor(document.getElementById('bclr_1').value)
				 && validColor(document.getElementById('pclr_1').value)
				 && validColor(document.getElementById('bclr_2').value)
				 && validColor(document.getElementById('pclr_2').value); 
	
	if ( parseInt(document.getElementById('c_min').value,10) > parseInt(document.getElementById('c_max').value, 10) )  {  
		res = false;
	}		
  return res;              
}


// start task (research) 
function startTask(e)   
{      			
 
  if ( setPar() == true ) 
	{	
    
		document.getElementById('no_task').style.display= 'none';
		
		initLayout(); 	
		
		stimGen();
		
		is_task= true;   // if ( timer_enabled ) finally "activate" timer 
		
		drawStim();    // draw first item (stimulus)
		
	}
	else  {  // do nothing. wait for user to change settings
		document.getElementById('no_task').style.display= 'block';
	}
}


// init DOM/html elements (positions,visibility,..)
function initLayout()
{	
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
	document.getElementById('task_section').style.display= 'block'; 	  	
	
 	document.getElementById('progress').style.display= 'block';   //document.getElementById('progress').setAttribute('style', 'width:100%');	 
	if ( document.getElementById('prg_type').value == '2' || ( document.getElementById('prg_type').value == '1' && timer_enabled == false ) ) {
		document.getElementById('progress_holder').style.display= 'none'; 
	}	
	else  
	{
		document.getElementById('progress_holder').style.display= 'block';     
		//document.getElementById('progress').setAttribute('style', 'width:100%');   		
		if (document.getElementById('prg_type').value == '0') {
			document.getElementById('progress').className='progress-bar progress-bar-info progress-bar-striped';     //  items / remaining items
		}
		else if (document.getElementById('prg_type').value == '1')  {
			document.getElementById('progress').className= 'progress-bar progress-bar-danger progress-bar-striped';  // remaining time   
		} 	
		document.getElementById('progress').setAttribute('style', 'width:0');  // reset progress bar width
	}					
	document.getElementById('nav_items').textContent='';  // always reset  
  //document.getElementById('feedback').textContent='';  // in drawStim()
	
	document.getElementById('demo_div').style.display= demo ? 'block' : 'none';   // Check button visible or not
	
  setCanvas();
	
}  


// set html 2 canvas element absolute positions
function setCanvas()
{
	// canvas must be after set footer content => progress_holder visible or not	
	let y = innerHeight - document.getElementById('footer').offsetHeight;
	document.getElementById('footer').top= y;
	
	let c= document.getElementById('can_2'), ctx= c.getContext('2d');
	
	if (stype)  {
		c.style.display= 'none';
	}
	else  {
			// set can_2 in top/upper part of screen. Only visible if Solution type - accepted => Partial = '2' = false
		
		c.style.display = 'block';   

		c.top= 0; 
		c.style.top = '0px';	
		
		c.width = innerWidth;  
		c.style.width = c.width + 'px';	     
		
		c.height = y/2;   
		c.style.height = c.height + 'px';  
	}
	ctx.textAlign = 'center';    // ctx.textAlign = "left" || "right" || "center" || "start" || "end";   Default value: 	start
	ctx.textBaseline = 'middle';  // const baselines = ['top', 'hanging', 'middle', 'alphabetic', 'ideographic', 'bottom'];  Default value: alphabetic
		
	ctx.lineWidth= 1; 
	//ctx.strokeStyle= pclr_2;  // for draw border in drawStim() => equal to fillStyle.
		
  c= document.getElementById('can_1'), ctx= c.getContext('2d'); // always is visible
	c.style.display = 'block';   
	
	c.left= 0;
	c.style.left = '0px';
	
	c.width = innerWidth;  
	c.style.width = c.width + 'px';
	
  if (stype)  {  
		c.top= 0; 
		c.style.top = '0px';	
		
		c.height = y;  // full screen height - footer height
		c.style.height = c.height + 'px';  
	}  
	else  {      
		c.top= y/2;  // set (bottom/down part of screen)	
		c.style.top = c.top + 'px';	   
		
		c.height = y/2;  // 0.5 x (full screen height - footer height)
		c.style.height = c.height + 'px';  
	}
	
	ctx.textAlign = 'center';    // ctx.textAlign = "left" || "right" || "center" || "start" || "end";   Default value: 	start
	ctx.textBaseline = 'middle';  // const baselines = ['top', 'hanging', 'middle', 'alphabetic', 'ideographic', 'bottom'];  Default value: alphabetic
		
	ctx.lineWidth= 1;
	//ctx.strokeStyle= pclr_1;  // for draw border in drawStim() => equal to fillStyle.
}


/* check internal application parameters (and set some global vars)
  return true if task can be started (or false if not)
  set content of id="no_task" validator message  */
function setPar()  
{  
	let ok = true;  // function result
	
	if ( c_min > c_max ) {  // this should be impossible but because of danger of indefinite loop in stimGen()
    let v = Math.min(c_min, c_max);
		c_min= v;
		c_max= v;
		document.getElementById('c_min').value= v;
		document.getElementById('c_max').value= v;
	}
	document.getElementById('no_task').innerHTML= '';
	
	if ( c_min < 2 || c_max < 2  || iter < 1 ) {
		document.getElementById('no_task').innerHTML= 'Minimum (and Maximum) number of letters used must be greater than 1. The number of iterations must be greater than 0.';
		ok= false;
	}

  if (	(c_max - c_min + 1) * iter > wor_cnt ) {
		document.getElementById('no_task').innerHTML+= '<br>The number of items is greater than the total number of words loaded from the dictionary.';
		ok= false;		
	}
	
	// Make sure there are words for all selected stimulus lengths (word count must be > 0) 
	let arr= [];  
	for (let len= c_min; len <= c_max; len++)  
	{ 
		if ( typeof words[len] === 'undefined' || Object.keys( words[len] ).length == 0 ) {
			arr.push(len);
		}
	}
	if ( arr.length > 0 )  { 
		document.getElementById('no_task').innerHTML += '<br>For these selected word lengths, the number of words in the loaded dictionary is zero: ' + arr.join(' ') + '.<br>You must have a minimum of one word loaded for each required stimulus length.'; 
		ok= false;
	}	
	
	// There are not enough words for the desired number of iterations. 	
	arr = [];
	for (let key in words)	
	{  // 1. key is word length
		if ( parseInt(key, 10) >= c_min && parseInt(key, 10) <= c_max )  
		{  
			if ( Object.keys( words[key] ).length < iter )  {    // check frequencies for each selected word length 
				arr.push(key);   // arr = array of problematic stimulus length
			}
		}
	}  
	if ( arr.length > 0 ) 
	{
		document.getElementById('no_task').innerHTML += '<br>There are not enough words in the loaded dictionary for these word lengths: <br>' + arr.join(' ');
		document.getElementById('no_task').innerHTML += '<br>You must reduce the number of iterations to a lower value.';
		ok= false;
	}
	
	if ( !stype && limen > Math.min(c_min, c_max) ) {
		document.getElementById('no_task').innerHTML += '<br>Minimum word length to be accepted as valid (threshold) is too big. The maximum allowed value is = ' + Math.min(c_min, c_max);
    ok= false;
	}		
	
	if ( ok )
	{ // init tsak
		item = 0; 
		counter = 0;       // reset timer counter		
		return true;
  }
	else  {
		document.getElementById('no_task').innerHTML += '<br>Unable to start task.'
		return false;
  }	
}


// generate all stimuli for each item in range 0..item_cnt-1	
function stimGen()	
{		
  stim = [];	    // stim = [items] x [character count in this item]   
	ans = [];
	cor = [];
		
	// col => number of displayed characters on screen (also canvas hidden column count)
	let col= Math.min(c_min, c_max), end= Math.max(c_min, c_max), sel= {};  
	while ( col <= end )  
	{			
    let len= col.toString();  // 1. key in words object is string length => len
		for (let i= 0; i< iter; i++) 
		{   
	    let txt, row= []; 			
			do  {
				txt = rndKey( words[len] );    // random word with length = len
				txt = txt.replace(/\s/g, '');  // Remove all Whitespace from a String
				//txt = txt.replace(/  +/g, ''); // Regex to replace multiple spaces with a single space
			}	while ( sel[txt] == 1 );
			sel[txt] = 1;  // remember this choice (sentence)	
			
	    // after selecting stimuli from dictionary lower(upper)Case txt (else indefinite loop danger) => duplicates are allowed for 2. key (aBC != abc =! Abc)			
			// string is not reversed from file. The text renders as it is.
			if ( is_lower == '2' )  {
				txt= txt.toLocaleLowerCase();   //txt.toLowerCase();
			}
			else if ( is_lower == '3' )  {
				txt= txt.toLocaleUpperCase();  //txt.toUpperCase(); 
			}				
					
			cor.push(txt);    // default correct anwswer for this item (note: can be more correct answers)	
			for (let j= 0, size= txt.length; j < size; j++)  {
				row.push(txt[j]);				
			}
			
			row = shuffle(row);  // always shuffle (on footer is (one of) correct solution if demo == true)
			
			stim.push(row);			
		}		
	
		col++;	
	} 	
		
	item_cnt= stim.length;   // set item_cnt => total items count in task 

	if (is_shuffle) {
		shuffleStimCor();    // at the same time shuffle both global arrays
	}
  
	// after [shuffle] set empty user answer for each item. ans array have equal dimensions like stim array
	if ( stype ) {  // one row: answer = stimulus
		ans = stim;   // ans[i] == stim[i]
	}
	else  
	{  // set empty string in each hidden canvas column (top row)
		for (let i = 0; i < item_cnt; i++) 
		{
			let len= stim[i].length, row= [];
			for (let j = 0; j < len; j++)  {
				row.push('');  
			}	
			ans.push(row);   // ans = [items] x [character count in this item]
		}
	}
}

 
// unbiased shuffle algorithm: Fisher-Yates (aka Knuth) Shuffle.
function shuffle(arr) 
{
  let cur_index = arr.length, tmp, rnd_index;
  // While there remain elements to shuffle...
  while (0 !== cur_index) {    
    rnd_index = Math.floor(Math.random() * cur_index);     // Pick a remaining element...
    cur_index--;   
    tmp = arr[cur_index];          // And swap it with the current element.
    arr[cur_index] = arr[rnd_index];
    arr[rnd_index] = tmp;
  }
  return arr;
}


/* shuffle global window stimulus and correct solution array => stim[] and cor[]
	at the same time shuffle both global arrays
	unbiased shuffle algorithm: Fisher-Yates (aka Knuth) Shuffle.  */
function shuffleStimCor() 
{  
	let cur_ind = stim.length, rnd_ind, tmp;
  // while there remain elements to shuffle...
  while (0 !== cur_ind) 
	{    
    rnd_ind = Math.floor(Math.random() * cur_ind);     // pick a remaining element...
    cur_ind--;    
    tmp = stim[cur_ind];          // swap it with the current element.
    stim[cur_ind] = stim[rnd_ind];
    stim[rnd_ind] = tmp;    
		// don't forget to shuffle correct answers
    tmp = cor[cur_ind];          // swap it with the current element.
    cor[cur_ind] = cor[rnd_ind];
    cor[rnd_ind] = tmp;  
  }
}


// draw (repaint) current stimulus           
function drawStim() 
{    		 
  let can = document.getElementById('can_1'), ctx = can.getContext('2d');  
	let col_cnt= stim[item].length;
	let width  = can.width / col_cnt;  // cell width => equal for id= 'can_1' and id= 'can_2'
	let height = can.height; // cell height => if can_2 is visible => equal for id= 'can_1' and id= 'can_2'
	let r = Math.floor( Math.min(width,height)/2 );  // const
	ctx.font = r.toString() + 'px ' + font_family;   	// update font
	ctx.fillStyle= bclr_1;  // background-color for can_1
  ctx.fillRect(0, 0, can.width, height);
  ctx.fillStyle= pclr_1;  // 'font-color' - for can_1
	if ( is_border ) {
		ctx.strokeStyle= pclr_1;  // for draw border only
	}

	let up= stype ? 0 : 1;  // id="can_2" is not repainted if stype == false 
	/* id= can_1 is always visible. can_2 is visible only if (!stype). in this case can_2 appears in top part (row) of screen
		 if (stype) only use can_1 => full screen => draw ans[]. it is init in stimGen(): ans[item]= stim[item] 
	   else for (!stype) => if i = 0 => draw bottom row stim[] arr. for i = 1 => draw top row ans[] array */
	for (let i = 0; i <= up; i++)
	{		
			if ( i == 1 ) {  // for (!stype) draw top row. change canvas object reference. for i = 1 => draw top row with ans[]
				can= document.getElementById('can_2');
				ctx = can.getContext('2d');	
        ctx.font = r.toString() + 'px ' + font_family;   	// must update font for different ctx	
        ctx.fillStyle= bclr_2;  // background-color for can_2
				ctx.fillRect(0, 0, can.width, can.height);
 			  ctx.fillStyle= pclr_2;  // 'font-color' for can_2
				if ( is_border ) {
					ctx.strokeStyle= pclr_2; 
				}
			}			
			for (let col= 0; col < col_cnt; col++)  
			{                						
				let left= col * width;   // cell left bound. width and height are equal for both canvas if can_2 is visible (can_1 and can_2). see: setCanvas()     
				let txt = i == 0 ? stim[item][col] : ans[item][col];	// if i = 0 => draw bottom row stim[] arr. for i = 1 => draw top row ans[] array							
				ctx.fillText(txt, Math.round(left+width/2), Math.round(height/2));	 
				if ( is_border )	{
					ctx.strokeRect(left, 0, width, height);	
				}			
			}
	} 

	updateFooter();	
	document.getElementById('correct').textContent= cor[item];
	document.getElementById('feedback').textContent= '';  // always reset
}


// update footer when show each stimulus in task. update buttons and update progress bar
function updateFooter()
{  
	document.getElementById('prior').style.display= item > 0 && is_back ? 'inline': 'none';  // set footer navigation
	let btn= document.getElementById('next');     
	if (item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update 'Next item' image on 'right arrow' button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Go forward';
	}						
  if (prg_type == '0' )  {    // update Progress bar => show current item count and remaining item count on Progress Bar                                                                                                                                                       
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent=  (item + 1) + ' / ' + (item_cnt - item-1);  
	}
}


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
}	


// Changes application settings that can be changed directly by the user within the options form
// These variables below can be directly modified by the user on the options form
function setVarPar()  
{   
	
	if (document.getElementById('c_min').validity.valid && document.getElementById('c_max').validity.valid 
    && (parseInt(document.getElementById('c_min').value,10) <= parseInt(document.getElementById('c_max').value, 10) ) )   {  
      c_min= parseInt(document.getElementById('c_min').value, 10); 
      c_max= parseInt(document.getElementById('c_max').value, 10); 
  }	

	if (document.getElementById('iter').validity.valid)  {  
	  iter = parseInt(document.getElementById('iter').value, 10);			
  } 	
	
	stype= document.getElementById('stype').value == '1' ? true : false;  // Solution type - accepted: Complete ('1', true) or Partial ('2', false)
	
	is_lower= document.getElementById('is_lower').value; 
	
	is_ltr= document.getElementById('is_ltr').value;
	
	if ( document.getElementById('limen').validity.valid ) {
		limen= parseInt(document.getElementById('limen').value, 10);
	}

  timer_enabled= document.getElementById('timer_enabled').checked ? true : false;  // timer is enabled or disabled	
	
	if (timer_enabled && document.getElementById('timer').validity.valid) {
		timer= Math.round(parseInt(document.getElementById('timer').value, 10));  // task time (s)
  }	
		
	font_family= document.getElementById('font_family').value;	
  
	is_back= document.getElementById('is_back').checked ? true : false; 
	
	prg_type= document.getElementById('prg_type').value;  
	
  is_shuffle= document.getElementById('is_shuffle').checked ? true : false;   
	
  demo= document.getElementById('demo').checked ? true : false;	

	if (validColor(document.getElementById('bclr_1').value))  {		
		bclr_1= document.getElementById('bclr_1').value;
  }
	
	if (validColor(document.getElementById('pclr_1').value))  {		
		pclr_1= document.getElementById('pclr_1').value;
  }
	
	if (validColor(document.getElementById('bclr_2').value))  {		
		bclr_2= document.getElementById('bclr_2').value;
  }
	
	if (validColor(document.getElementById('pclr_2').value))  {		
		pclr_2= document.getElementById('pclr_2').value;
  }
		
  is_border = document.getElementById('is_border').checked ? true : false;
	
  store= (is_storage && document.getElementById('store').value === '1') ? true : false;
 
}


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// goto previous item (stimulus)
function priorItem()  
{      
  if (item > 0)  {
    document.getElementById('can_m').style.display= 'none';  // raster can be changed in next/prior task
		item--;  
    drawStim();     // draw again previous stimulus
  }         
}


// goto next item (stimulus)
function nextItem()  
{    		
	item++;					
	document.getElementById('can_m').style.display= 'none'; // raster can be changed in next/prior task
	if (item < item_cnt)  {         		 
    drawStim();     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


/* set absolute positions (left, top, width, height) for can_m object	
   canvas => mousemove event	=> a pointing device is moved over an element (fired continuously as the mouse moves).
	 this event is fired by 2 html canvas elements => document.getElementById('can_1') and document.getElementById('can_1') */
function mouseMove(e)
{  
  if ( is_task ) 
	{   	
		let can = e.target || e.srcElement;  			// source canvas element object reference  		
		let dx =  can.width / stim[item].length;  // ex getCellFromXY()
		let col = Math.floor ( e.offsetX / dx );  // stimulus (character) array index
	
		// set absolute positions (left, top, width, height) for m_canvas object			
		let can_m = document.getElementById('can_m');
		can_m.style.left= col * dx  + 'px';  // top and left properties require a unit of measure, not just a number
		can_m.style.top= can.top + 'px';    
		can_m.width= dx;  
		can_m.height= can.height;  
  
		//if ( can_m.style.display != 'block' )  {  // ? performance
		  can_m.style.display= 'block';	
		//}	
		let data = [can.id, col];  //  xxx
		can_m.setAttribute('data-cell', data.join(';'));   // 'data-cell' = 'canvas.id;column index'  
				
		let ctx= can_m.getContext('2d'), width = can_m.width, height= can_m.height, txt;
		let r = Math.floor( Math.min(width,height)/2 );   
		ctx.font = r.toString() + 'px ' + font_family;   	// update font	
		
		ctx.textAlign = 'center';    // ctx.textAlign = "left" || "right" || "center" || "start" || "end";   Default value: 	start
	  ctx.textBaseline = 'middle';  // ctx.textBaseline = "top" || "hanging" || "middle" || "alphabetic" || "ideographic" || "bottom";    Default value: alphabetic
				
		if ( can.id == 'can_1' )  {	
		  txt = stim[item][col];
			ctx.fillStyle= bclr_1;
			ctx.fillRect(0, 0, width, height);    
			ctx.fillStyle= pclr_1;
		}
		else  {
			txt = ans[item][col];
			ctx.fillStyle= bclr_2;
			ctx.fillRect(0, 0, width, height); 
			ctx.fillStyle= pclr_2;			
		}	
		ctx.fillText(txt, Math.round(width/2), Math.round(height/2));					
	}	
}


// ondragstart event for html can_m object 
// html => <canvas id="can_m" draggable="true" ondragstart="dragStart(event)"></canvas> 
function dragStart(e)  {    
	// real MIME type "text/plain" being unsupported in IE11
	if ( is_task )  {  // data = [can.id, col]
		e.dataTransfer.setData('text', document.getElementById('can_m').getAttribute('data-cell') );	//  [0] = canvas.id; [1]= cell index
	}
}


// ondragover() event for html canvas element/object
// the ondragover event specifies wher the dragged data can be dropped
// html:  <canvas id="can_1" onmousemove="mouseMove(event)" ondragover="allowDrop(event)" ondrop="drop(event)"></canvas> /
function allowDrop(e)  {
  if ( is_task )  {
		e.preventDefault();
	}
}


// ondrop() event for html canvas element/object
// html: <canvas id="can_1" onmousemove="mouseMove(event)" ondragover="allowDrop(event)" ondrop="drop(event)"></canvas> 
function drop(e)  
{  
  if ( is_task )
	{
		e.preventDefault();   
		
				// data[0] => id of source canvas element; data[1] => source cell column index		
		    // real MIME type "text/plain" being unsupported in IE11
		let data= e.dataTransfer.getData('text').split(';');  //  data from start or SOURCE rect => source cell position 		
		let s_col= data[1];  // source cell index		
		
		let d_can = e.target || e.srcElement;  			// destination canvas element object reference  			
		let d_col = Math.floor ( e.offsetX / ( d_can.width / stim[item].length ) );  // destination character array index
				
		// update (swap) user answer => swap rectangles/cells positions
		if ( stype && data[0] == 'can_1' && d_can.id == 'can_1' )  {  // Down row => Down row drag & drop		
		  let tmp= ans[item][s_col];
			ans[item][s_col]= ans[item][d_col];
			ans[item][d_col]= tmp;
		}
		else if ( data[0] == 'can_1' && d_can.id == 'can_2' ) {  // Down row => Up row drag & drop = stim <=> ans
			let tmp= stim[item][s_col];
			stim[item][s_col]= ans[item][d_col];
			ans[item][d_col]= tmp;
		}
		else if ( data[0] == 'can_2' && d_can.id == 'can_1' ) {  // Up => Down drag & drop = ans <=> stim
			let tmp= ans[item][s_col];
			ans[item][s_col]= stim[item][d_col];
			stim[item][d_col]= tmp;
		}
		else if ( data[0] == 'can_2' && d_can.id == 'can_2' && s_col != d_col ) {  // Up => Up drag & drop = ans <=> ans
			let tmp= ans[item][s_col];
			ans[item][s_col]= ans[item][d_col];
			ans[item][d_col]= tmp;
		}

		drawStim();  // repaint stimulus
	}
}


// https://developer.mozilla.org/en-US/docs/Web/API/FileReader/readAsText
// text html: https://www.howtoexcel.org/split-text-by-delimiter/
// Loads the user's dictionary (stimuli) from a text file
function loadStimuli() {
	let file = document.getElementById('file_dic').files[0];
	let delimit = document.getElementById('newline').checked ? '\n' : document.getElementById('delim').value;
	let reader = new FileReader(), lines, cnt= 0, dup= 0, sht= 0, eql= 0;
	reader.addEventListener('load', function () {	
		//content.innerText = reader.result; // this will then display a text file
		words= {}; 
		lines = reader.result.split(delimit);  // xxx new array (reader.result == a complete text file) 
		for  (let i = 0, tot= lines.length; i < tot; i++)
		{      				
			let txt = lines[i].toString().replace(/\r\n|\n|\r/g, '').trim();  			
			let len= txt.length; // len (word length) => is 1. key in words object
			if ( len > 1 )  
			{   // escape string length < 2
				let same = /^(.)\1*$/.test(txt);  // https://stackoverflow.com/questions/3258444/how-can-i-write-the-regex-all-characters-are-the-same
				if ( !same ) // escape if all characters are the same in string 
				{
					if ( typeof words[len] === 'undefined' ) {
						words[len] = {};
					}
					if ( words[len][txt]  )  {
						dup++;  // escape duplicates
					}
					else  {
						words[len][txt] = 1;
						cnt++;
					}
				}
				else  {
					eql++;
				}
			}
			else  {
				sht++;
			}
		}		
 
		dic_def= false;   // default Dictionary is not used => user loaded custom Dictionary loaded from txt file 
   	displayFreq();	
		
		document.getElementById('freq_rep').innerHTML+= '<br><hr><br>Uploaded file = ' + file.name + '<br>Number of unloaded too short words (all those with less than 2 characters) = ' + sht + '<br>Number of unloaded trivial words that have the same characters (bbbbb, aaa, hhhhh) = ' + eql + '<br>Number of words repeated = ' + dup + '<br>Total number of successfully loaded words = ' + cnt;  
		alert('File upload complete: ' + file.name + ' \n Total number of successfully loaded words = ' + cnt + ' \n Number of unloaded too short words (all those with less than 2 characters) = ' + sht + ' \n Number of unloaded trivial words that have the same characters (bbbbb, aaa, hhhhh) = '  + eql + ' \n Number of words repeated = ' + dup);   // default Dictionary is not used => user loaded custom Dictionary loaded from txt file    
		//document.getElementById('my_div').style.display= 'block';
	}, false);
	/*reader.addEventListener('error', function() {
		alert('Error occurred reading file: ' + file.name);
	});*/

	if (file) {
		reader.readAsText(file);
	}
}


// https://www.freecodecamp.org/news/how-to-reverse-a-string-in-javascript-in-3-different-ways-75e4763c68cb/
function reverseString(str) {
    return str.split("").reverse().join("");
}


// check if created word is valid (only when demo == true)
function checkWord()
{
	let txt = ans[item].join(''), e = document.getElementById('feedback'); 
	if ( txt.length == 0 )  {
		e.textContent= "You didn't create a word.";
    if (!stype) {
			e.textContent +=	' Move the letters from the bottom line to the top line so that they form a valid word.';
	  }
	}
	else  {
		let res = getScore(item);
		if ( is_ltr == '2' ) {
			txt= reverseString(txt);
		}
		e.textContent=  'Word check: "' + txt + '". ';
		e.textContent+=  res > 0 ? 'The word is found in the used dictionary. Number of points: ' + res  + '.' : 'The word is invalid. Number of points: 0.';
		if (!stype) {
			e.textContent+= ' Minimum validity threshold = ' + limen + '.';
		}
	}
}	


// return number of points for item i
function getScore(i) 
{
	let treshold = stype ? ans[i].length : limen;    // if (stype) => ans = stim. see stimGen()
	let txt= ans[i].join(''), len= txt.length, txt_low= txt.toLocaleLowerCase(), txt_up= txt.toLocaleUpperCase(), r_txt, r_txt_low, r_txt_up, ok;
	if ( is_ltr != '1' )  {  // getScore() will do the job correctly if ltr != '1'
		r_txt= reverseString(txt), r_txt_low= r_txt.toLocaleLowerCase(), r_txt_up= r_txt.toLocaleUpperCase();
	}
	switch ( is_ltr )  {	 
		case '1':  // parse left-to-right
			ok = len >= treshold && ( words[len][txt] == 1 || words[len][txt_low] == 1 || words[len][txt_up] == 1 );
	    break;
		case '2':	// parse right-to-left
			ok = len >= treshold && ( words[len][r_txt] == 1 || words[len][r_txt_low] == 1 || words[len][r_txt_up] == 1 );
			break;
		case '3':	// parse both directions: left OR right
			ok = len >= treshold && ( words[len][txt] == 1 || words[len][txt_low] == 1 || words[len][txt_up] == 1 );
			if ( !ok ) {
				ok = len >= treshold && ( words[len][r_txt] == 1 || words[len][r_txt_low] == 1 || words[len][r_txt_up] == 1 );
			}
			break;	
	}	
	let res = ok == true ? len : 0;
	return res;
}	


// end of task (research)
function theEnd()  
{    
	is_task = false;

	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
  document.getElementById('can_1').style.display= 'none';  
  document.getElementById('can_2').style.display= 'none';  
	document.getElementById('can_m').style.display= 'none';	
	document.getElementById('result_wrapper').style.display= 'block';
  
	let header= [ 'Item', 'Characters', 'Answer', 'Points', '*Solution' ];
	let size= stim.length,  score = 0, max_score = 0, arr = [], rows = [], len, p, s;		
	for (let i = 0; i < size; i++)  {		// for (let i = 0; i < item_cnt; i++)			
		p = getScore(i);   // number of points in this i-item	
		arr.push(p);
		score+= p;  // sum of points in whole task
		len= stim[i].length;
		max_score+= len;
		s= ans[i].join('');  // user answer for this i-item	  
		if ( is_ltr == '2' ) {
			s= reverseString(s);
		}
		rows[i]= [ parseInt(i+1,10), len, s, p, cor[i] ]; 
	}  
	rows.push (  [ 'Sum',  max_score, '', score, '' ] );   // totals 	
  createTableArray(document.getElementById('table1'), rows, header); 
  document.getElementById('s_limen').textContent= stype ? ' number of characters displayed' : ' '  + limen;  
	
	// table2 => Descriptive statistics	
	const D = 3;  // number of decimals
	rows = [];  // reset table rows (for table2)
	let n = arr.length;  
	let m =   n > 0  ?  mean(arr).toFixed(D) : '';      
	let sd =  n > 1  ?  stDev(arr, m).toFixed(D) : '';   	 	
  rows.push( [ m, sd, n ] );  // last row
	
	header= ['Average', 'Standard deviation', 'N']; // header		
	createTableArray(document.getElementById('table2'), rows, header); 

  // update 'score' label => above "Repeat" button
  let percent = (100 * score / max_score).toFixed(D);
	document.getElementById('score').textContent= 'Score= ' + score + '. Total number of characters displayed= ' + max_score + '. Percent= ' + percent + '%';
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
	
	try 
  {
    if (store) { 
			setStorage(score, max_score, percent, m, sd);   			
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }  
}


// return arithmetic mean (mean, average) of input array (arr)
// no validation => input array must be in valid number format	
function mean(arr)  
{	
	let len= arr.length, sum= 0;
	for (let i=0; i<len; i++) {
		sum+= arr[i];
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation (with Bessel's correction) => N-1
// no validation => input array must be in valid number format
// arr = array argument;  mean = average for input arr 
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
		sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	if (dd < 10) { dd='0'+dd; } 
	if (mm < 10) { mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 score = sum (number) of points in task
   max_score = maximum possible points in task
   percent = 100 * points/ max_points	 
   mean = average (mean) for user points 
   sd = standard deviation for user points	 */   
function setStorage(score, max_score, percent, mean, sd)  
{    
	let oldItems = JSON.parse(localStorage.getItem('lpt-2-v1.0')) || []; 
  let id= oldItems.length + 1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   
			'score' : score,
			'max-score' : max_score,
			'percent' : percent,
			'mean'  : mean, 	
			'sd' : sd,			
			'letter-min' : c_min,
      'letter-max' : c_max,			
			'iteration' : iter,
			'items' : item_cnt,
			'solution' : document.getElementById('stype').value,			
			'letter-type' : is_lower,
			'parse' : is_ltr,
			'threshold' : stype ? '' : limen,	
			'timer' : timer_enabled ? timer : '',	
			'is-back' : is_back ? 1 : 0,
			'shuffle' : is_shuffle ? 1 : 0,
			'demo' : demo ? 1 : 0,
      'default-dic' : dic_def  ? 1 : 0			
  };

  oldItems.push(newItem);

  localStorage.setItem('lpt-2-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
          col.push(key);
        }
    }
  } 
  let table = document.createElement('table');
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement('th');      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= '';
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('lpt-2-v1.0')  ?  JSON.parse(localStorage.getItem('lpt-2-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    //alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('lpt-2-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('lpt-2-v1.0');
	if (!data)  {
    //alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('lpt-2-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object (div or table)
function newWinLast(id)
{ 	
	let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
    case 1:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  
			break;
	}		
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
